#pragma once

#include <memory>

// A map from `K`s to `V`s, represented as a binary search tree.
template <class K, class V>
class Bst_map
{
public:
    size_t size() const;
    bool empty() const;

    V const* lookup(K const& key) const;
    V* lookup(K const& key);

    void insert(K const& key, V const& value);

private:
    struct Node_;
    using link_t = std::unique_ptr<Node_>;

    struct Node_
    {
        Node_(K const& key, V const& value)
                : key(key), value(value)
        { }

        K const key;
        V       value;
        link_t  left;
        link_t  right;
    };

    link_t root_;
    size_t size_ = 0;
};

template<class K, class V>
size_t Bst_map<K, V>::size() const
{
    return size_;
}

template<class K, class V>
bool Bst_map<K, V>::empty() const
{
    return size_ == 0;
}

template<class K, class V>
V const* Bst_map<K, V>::lookup(K const& key) const
{
    Node_ const* current = root_.get();

    while (current) {
        if (key < current->key)
            current = current->left.get();
        else if (key > current->key)
            current = current->right.get();
        else
            return &current->value;
    }

    return nullptr;
}

template<class K, class V>
V* Bst_map<K, V>::lookup(K const& key)
{
    Node_* current = root_.get();

    while (current) {
        if (key < current->key)
            current = current->left.get();
        else if (key > current->key)
            current = current->right.get();
        else
            return &current->value;
    }

    return nullptr;
}

template<class K, class V>
void Bst_map<K, V>::insert(K const& key, V const& value)
{
    link_t* current = &root_;

    while (*current) {
        if (key < (*current)->key) {
            current = &(*current)->left;
        } else if (key > (*current)->key) {
            current = &(*current)->right;
        } else {
            (*current)->value = value;
            return;
        }
    }

    *current = std::make_unique<Node_>(key, value);
    ++size_;
}

